// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

var DotNetEntropyLib = {
    $DOTNETENTROPY: {
        // batchedQuotaMax is the max number of bytes as specified by the api spec.
        // If the byteLength of array is greater than 65536, throw a QuotaExceededError and terminate the algorithm.
        // https://www.w3.org/TR/WebCryptoAPI/#Crypto-method-getRandomValues
        batchedQuotaMax: 65536,
        getBatchedRandomValues: function (buffer, bufferLength) {
            // for modern web browsers
            // map the work array to the memory buffer passed with the length
            for (var i = 0; i < bufferLength; i += this.batchedQuotaMax) {
                var view = new Uint8Array(Module.HEAPU8.buffer, buffer + i, Math.min(bufferLength - i, this.batchedQuotaMax));
                crypto.getRandomValues(view)
            }
        }
    },
    dotnet_browser_entropy : function (buffer, bufferLength) {
        // check that we have crypto available
        if (typeof crypto === 'object' && typeof crypto['getRandomValues'] === 'function') {
            DOTNETENTROPY.getBatchedRandomValues(buffer, bufferLength)
            return 0;
        } else {
            // we couldn't find a proper implementation, as Math.random() is not suitable
            // instead of aborting here we will return and let managed code handle the message
            return -1;
        }
    },
};

autoAddDeps(DotNetEntropyLib, '$DOTNETENTROPY')
mergeInto(LibraryManager.library, DotNetEntropyLib)

// SIG // Begin signature block
// SIG // MIIoiwYJKoZIhvcNAQcCoIIofDCCKHgCAQExDzANBglg
// SIG // hkgBZQMEAgEFADB3BgorBgEEAYI3AgEEoGkwZzAyBgor
// SIG // BgEEAYI3AgEeMCQCAQEEEBDgyQbOONQRoqMAEEvTUJAC
// SIG // AQACAQACAQACAQACAQAwMTANBglghkgBZQMEAgEFAAQg
// SIG // He0Yy7K+G59QXMN5sHHtOtGJyoFg0Ie3YpqLi1MNoKeg
// SIG // gg3wMIIGbjCCBFagAwIBAgITMwAAAo1+R8OCfgUaKgAA
// SIG // AAACjTANBgkqhkiG9w0BAQwFADB+MQswCQYDVQQGEwJV
// SIG // UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
// SIG // UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
// SIG // cmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBT
// SIG // aWduaW5nIFBDQSAyMDExMB4XDTIxMTAxNDE4NDUxNFoX
// SIG // DTIyMTAxMzE4NDUxNFowYzELMAkGA1UEBhMCVVMxEzAR
// SIG // BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
// SIG // bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
// SIG // bjENMAsGA1UEAxMELk5FVDCCAaIwDQYJKoZIhvcNAQEB
// SIG // BQADggGPADCCAYoCggGBAM+nYwbxkHhF3CQTxhfbfq0y
// SIG // Y9iNmf+vpsXyHr+W14sNKW2VmN48wwUttFgkElZWXDR7
// SIG // /LVrKRjN1wUWy/bzsFToydMsiIzNT1HUivMfeT/cykpT
// SIG // N/cVL/ZvvGrnhJeXQEn1xrnGNqW3ps0NjQQLPd2fvIy1
// SIG // Y/YAIh9r2+dHkYj+VjmEtv9v7r2jbtklWw6OFgOwkB8f
// SIG // GA+15Qiny+1dE5WvItLj/DGrPmCWz4MVgfG42ntE481F
// SIG // Ly4U74rBEDtaNahOtPUSS8yTjUeNIgi3eTkznStetnjg
// SIG // r+Bn0Io4KhMqkwA7cav5wxlORTU/OTdM6PVJrw6NKC6I
// SIG // ztKqeOjlFs26h1c5eBY6ZKIbBwNkDQuSq/P52gOjsTzh
// SIG // /s+9JPwbXzr/plrAXIXZh178HTrsr5gP9iaPXWIMDvlM
// SIG // Fw54saZB68Hh+D1XiAKmOvct4etdk8v8wlJ96O3j8S2o
// SIG // omSdqcALeycc7hVnpJ8j6hFVW9hXFRqSb9VYn18cMu5u
// SIG // 3WvIkQIDAQABo4IBfjCCAXowHwYDVR0lBBgwFgYKKwYB
// SIG // BAGCN0wIAQYIKwYBBQUHAwMwHQYDVR0OBBYEFB4HrzFI
// SIG // RagJ4H8x6Jocx6igXl7OMFAGA1UdEQRJMEekRTBDMSkw
// SIG // JwYDVQQLEyBNaWNyb3NvZnQgT3BlcmF0aW9ucyBQdWVy
// SIG // dG8gUmljbzEWMBQGA1UEBRMNNDY0MjIzKzQ2ODYyNjAf
// SIG // BgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBU
// SIG // BgNVHR8ETTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jv
// SIG // c29mdC5jb20vcGtpb3BzL2NybC9NaWNDb2RTaWdQQ0Ey
// SIG // MDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUw
// SIG // UzBRBggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3Nv
// SIG // ZnQuY29tL3BraW9wcy9jZXJ0cy9NaWNDb2RTaWdQQ0Ey
// SIG // MDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
// SIG // DQYJKoZIhvcNAQEMBQADggIBAB4qmkYG7kKK3A6/oZNe
// SIG // IP9JhNg7SX+VnacQGuwIHW2TxICObVUVh7Pq8m+xG9Ec
// SIG // o4Wl8AoArhOWnp3IMWFiF+vxGD7zaJpG77kxFXDewsA8
// SIG // PnehwnMfHq6TliI5/65+FZB4Kf5Ey16s2Qk6nTSq/bsg
// SIG // T572aCkU9hPd5WXukhRfuQOnWn6lRWREhcqAReuFmik5
// SIG // YD+hgJZgo3sCDc01hVEgOIdwgjXMENALrAgaQlp/QFRX
// SIG // +DMRpW96eyFoKFRWiRudBhtSqf9I+WmTgzK9QStgT8mn
// SIG // njaY70f8/dcqs0nv4wrWb438wT1xddyIrQXMnObYZCqb
// SIG // 7JDNTPfRpKpfAykwhRmAJDDvDn/zNmlz/vcaU4+WLtBV
// SIG // 2zpyk4oVcZzJgMWgGl3gdg8+fNAcLoQwfRqk+wYJccu+
// SIG // IX8lR0h+CygomPKALmxSb2ShJsU3BXXd6E135PgCkPsv
// SIG // x3ntyeorbcAshUOIaqJamTOdWkNf5X97QoTDEuPsS2tI
// SIG // zI3munvtDZ14nykyYjf4eX8NR6pAwOEgMrWQ14taSKq6
// SIG // MaXNucGaqCzFw/L+4p115iZbOo69+OuOhbVNB2tIZjeK
// SIG // YE7QKKU+lAdzgZUacya+Mg1Ku3ndGdvDB8IT735c3nU3
// SIG // 8LV8Ytut5jxvaiA1om3DNumfVNAITHgnJF8p7x1DzIA5
// SIG // Nax2MIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkq
// SIG // hkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNV
// SIG // BAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
// SIG // HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEy
// SIG // MDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlmaWNh
// SIG // dGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4MjA1OTA5
// SIG // WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzET
// SIG // MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
// SIG // bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
// SIG // aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWdu
// SIG // aW5nIFBDQSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOC
// SIG // Ag8AMIICCgKCAgEAq/D6chAcLq3YbqqCEE00uvK2WCGf
// SIG // Qhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDI
// SIG // OdUBFDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv
// SIG // 2akrrnoJr9eWWcpgGgXpZnboMlImEi/nqwhQz7NEt13Y
// SIG // xC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
// SIG // +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy
// SIG // 4BI6t0le2O3tQ5GD2Xuye4Yb2T6xjF3oiU+EGvKhL1nk
// SIG // kDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEhNSXD
// SIG // OW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVf
// SIG // GCi2zCcoOCWYOUo2z3yxkq4cI6epZuxhH2rhKEmdX4ji
// SIG // JV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3s/gA4bys
// SIG // AoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLn
// SIG // JN+/RpnF78IcV9uDjexNSTCnq47f7Fufr/zdsGbiwZeB
// SIG // e+3W7UvnSSmnEyimp31ngOaKYnhfsi+E11ecXL93KCjx
// SIG // 7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwT
// SIG // XWCVmj5PM4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEA
// SIG // AaOCAe0wggHpMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1Ud
// SIG // DgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEE
// SIG // AYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYw
// SIG // DwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBRyLToC
// SIG // MZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
// SIG // hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2Ny
// SIG // bC9wcm9kdWN0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFf
// SIG // MDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggrBgEF
// SIG // BQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3Br
// SIG // aS9jZXJ0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNf
// SIG // MjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsGAQQBgjcu
// SIG // AzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNy
// SIG // b3NvZnQuY29tL3BraW9wcy9kb2NzL3ByaW1hcnljcHMu
// SIG // aHRtMEAGCCsGAQUFBwICMDQeMiAdAEwAZQBnAGEAbABf
// SIG // AHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQA
// SIG // LiAdMA0GCSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou
// SIG // 09h0ZyKbC5YR4WOSmUKWfdJ5DJDBZV8uLD74w3LRbYP+
// SIG // vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzS
// SIG // Gksc8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlE
// SIG // PXh6I/MTfaaQdION9MsmAkYqwooQu6SpBQyb7Wj6aC6V
// SIG // oCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
// SIG // 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560
// SIG // STkKxgrCxq2u5bLZ2xWIUUVYODJxJxp/sfQn+N4sOiBp
// SIG // mLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXhj38c
// SIG // bxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HN
// SIG // pZfQ7l1jQeNbB5yHPgZ3BtEGsXUfFL5hYbXw3MYbBL7f
// SIG // QccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtIEJmAH9AA
// SIG // KcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3
// SIG // S5fi63V4GuzqN5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8
// SIG // MO0ETI7f33VtY5E90Z1WTk+/gFcioXgRMiF670EKsT/7
// SIG // qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJ
// SIG // UnMTDXpQzTGCGfMwghnvAgEBMIGVMH4xCzAJBgNVBAYT
// SIG // AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
// SIG // EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29y
// SIG // cG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2Rl
// SIG // IFNpZ25pbmcgUENBIDIwMTECEzMAAAKNfkfDgn4FGioA
// SIG // AAAAAo0wDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
// SIG // AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEO
// SIG // MAwGCisGAQQBgjcCARUwLwYJKoZIhvcNAQkEMSIEIKTQ
// SIG // ts56jQmXdTOrJhxHePDRP6IYHMqtdIFdJu6K2KWDMEIG
// SIG // CisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBv
// SIG // AGYAdKEagBhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20w
// SIG // DQYJKoZIhvcNAQEBBQAEggGAwSBbzySO/VW3vahv1T0y
// SIG // 88ip5FqaP9kuuI7N4fxx0xW12HydKCwjNmtxoTIkaMEx
// SIG // JEu0SBxfu5xUJbQyOorn92EeT2KupsPKez734PFQLeZN
// SIG // n2byVhiYPrxalZ/XCkM0mjKsgYId46xwZF9hqnhdGXsY
// SIG // BTGubTgAuByETBmgyi0qpCs2jlMomNHUFAR95vqhJc0j
// SIG // F8LpAgBEs7EeYXxR6CrmUhMzC07OPBnvwagHVQfTLkg2
// SIG // yzLqyfv/g2/4/JUrzZ7GtKRWcolGj4PM7ErE4LECxuek
// SIG // Gp/HdGsawRgQJtc7XT+pvNW7iZ9VTXSeIKsnHbqbar/9
// SIG // +fWESU4Gx7SnrzzmIn9B8CvRTtKFEIhaWqxd+GNxndn7
// SIG // E61+GbpBx6+LkHjpMSLvPh+xLdiDhAnYf3hdXD4Rq6uc
// SIG // hVItAm7Jh39JPjoq5/hzh/PRsgbWLIfxLKxTP0ilihjF
// SIG // 2NH2nFHIlrY18XXGW8+C9odYdtLSMNipIuhD9uJYoXvE
// SIG // GwLO0CBCoYIW/TCCFvkGCisGAQQBgjcDAwExghbpMIIW
// SIG // 5QYJKoZIhvcNAQcCoIIW1jCCFtICAQMxDzANBglghkgB
// SIG // ZQMEAgEFADCCAVEGCyqGSIb3DQEJEAEEoIIBQASCATww
// SIG // ggE4AgEBBgorBgEEAYRZCgMBMDEwDQYJYIZIAWUDBAIB
// SIG // BQAEIHR0LZkXgJcUUCaXwS6t/QTuaZG6WwNcjyQnt0+u
// SIG // Z+NGAgZiYb2L/+sYEzIwMjIwNDI1MTkxNzE4LjU0OFow
// SIG // BIACAfSggdCkgc0wgcoxCzAJBgNVBAYTAlVTMRMwEQYD
// SIG // VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
// SIG // MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
// SIG // JTAjBgNVBAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJh
// SIG // dGlvbnMxJjAkBgNVBAsTHVRoYWxlcyBUU1MgRVNOOjNF
// SIG // N0EtRTM1OS1BMjVEMSUwIwYDVQQDExxNaWNyb3NvZnQg
// SIG // VGltZS1TdGFtcCBTZXJ2aWNloIIRVDCCBwwwggT0oAMC
// SIG // AQICEzMAAAGg6buMuw6i0XoAAQAAAaAwDQYJKoZIhvcN
// SIG // AQELBQAwfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldh
// SIG // c2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNV
// SIG // BAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UE
// SIG // AxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAw
// SIG // HhcNMjExMjAyMTkwNTIzWhcNMjMwMjI4MTkwNTIzWjCB
// SIG // yjELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
// SIG // b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
// SIG // Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWlj
// SIG // cm9zb2Z0IEFtZXJpY2EgT3BlcmF0aW9uczEmMCQGA1UE
// SIG // CxMdVGhhbGVzIFRTUyBFU046M0U3QS1FMzU5LUEyNUQx
// SIG // JTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNl
// SIG // cnZpY2UwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIK
// SIG // AoICAQC/2uIOaHGdAOj2YvhhI6C8iFAq7wrl/5WpPjj0
// SIG // fEHCi6Ivx/I02Jss/HVhkfGTMGttR5jRhhrJXydWDnOm
// SIG // zRU3B4G525T7pwkFNFBXumM/98l5k0U2XiaZ+bulXHe5
// SIG // 4x6uj/6v5VGFv+0Hh1dyjGUTPaREwS7x98Te5tFHEimP
// SIG // a+AsG2mM+n9NwfQRjd1LiECbcCZFkgwbliQ/akiMr1tZ
// SIG // mjkDbxtu2aQcXjEfDna8JH+wZmfdu0X7k6dJ5WGRFwzZ
// SIG // iLOJW4QhAEpeh2c1mmbtAfBnhSPN+E5yULfpfTT2wX8R
// SIG // bH6XfAg6sZx8896xq0+gUD9mHy8ZtpdEeE1ZA0HgByDW
// SIG // 2rJCbTAJAht71B7Rz2pPQmg5R3+vSCri8BecSB+Z8mwY
// SIG // L3uOS3R6beUBJ7iE4rPS9WC1w1fZR7K44ZSme2dI+O9/
// SIG // nhgb3MLYgm6zx3HhtLoGhGVPL+WoDkMnt93IGoO6kNBC
// SIG // M2X+Cs22ql2tPjkIRyxwxF6RsXh/QHnhKJgBzfO+e84I
// SIG // 3TYbI0i29zATL6yHOv5sEs1zaNMih27IwfWg4Q7+40L7
// SIG // e68uC6yD8EUEpaD2s2T59NhSauTzCEnAp5YrSscc9MQV
// SIG // Ii7g+5GAdC8pCv+0iRa7QIvalU+9lWgkyABU/niFHWPj
// SIG // yGoB4x3Kzo3tXB6aC3yZ/dTRXpJnaQIDAQABo4IBNjCC
// SIG // ATIwHQYDVR0OBBYEFHK5LlDYKU6RuJFsFC9EzwthjNDo
// SIG // MB8GA1UdIwQYMBaAFJ+nFV0AXmJdg/Tl0mWnG1M1Gely
// SIG // MF8GA1UdHwRYMFYwVKBSoFCGTmh0dHA6Ly93d3cubWlj
// SIG // cm9zb2Z0LmNvbS9wa2lvcHMvY3JsL01pY3Jvc29mdCUy
// SIG // MFRpbWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNybDBs
// SIG // BggrBgEFBQcBAQRgMF4wXAYIKwYBBQUHMAKGUGh0dHA6
// SIG // Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvY2VydHMv
// SIG // TWljcm9zb2Z0JTIwVGltZS1TdGFtcCUyMFBDQSUyMDIw
// SIG // MTAoMSkuY3J0MAwGA1UdEwEB/wQCMAAwEwYDVR0lBAww
// SIG // CgYIKwYBBQUHAwgwDQYJKoZIhvcNAQELBQADggIBADF9
// SIG // xgKr+N+slAmlbcEqQBlpL5PfBMqcLkS6ySeGJjG+LKX3
// SIG // Wov5pygrhKftXZ90NYWUftIZpzdYs4ehR5RlaE3eYubW
// SIG // lcNlwsKkcrGSDJKawbbDGfvO4h/1L13sg66hPib67mG9
// SIG // 6CAqRVF0c5MA1wiKjjl/5gfrbdNLHgtREQ8zCpbK4+66
// SIG // l1Fd0up9mxcOEEphhJr8U3whwFwoK+QJ/kxWogGtfDia
// SIG // q6RyoFWhP8uKSLVDV+MTETHZb3p2OwnBWE1W6071XDKd
// SIG // xRkN/pAEZ15E1LJNv9iYo1l1P/RdF+IzpMLGDAf/PlVv
// SIG // TUw3VrH9uaqbYr+rRxti+bM3ab1wv9v3xRLc+wPoniSx
// SIG // W2p69DN4Wo96IDFZIkLR+HcWCiqHVwFXngkCUfdMe3xm
// SIG // vOIXYRkTK0P6wPLfC+Os7oeVReMj2TA1QMMkgZ+rhPO0
// SIG // 7iW7N57zABvMiHJQdHRMeK3FBgR4faEvTjUAdKRQkKFV
// SIG // 82uE7w0UMnseJfX7ELDY9T4aWx2qwEqam9l7GHX4A2Zm
// SIG // 0nn1oaa/YxczJ7gIVERSGSOWLwEMxcFqBGPm9QSQ7ogM
// SIG // Bn5WHwkdTTkmanBb/Z2cDpxBxd1vOjyIm4BOFlLjB4pi
// SIG // vClO2ZksWKH7qBYloYa07U1O3C8jtbzGUdHyLCaVGBV8
// SIG // DfD5h8eOnyjraBG7PNNZMIIHcTCCBVmgAwIBAgITMwAA
// SIG // ABXF52ueAptJmQAAAAAAFTANBgkqhkiG9w0BAQsFADCB
// SIG // iDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
// SIG // b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
// SIG // Y3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
// SIG // cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5
// SIG // IDIwMTAwHhcNMjEwOTMwMTgyMjI1WhcNMzAwOTMwMTgz
// SIG // MjI1WjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
// SIG // aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
// SIG // ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQD
// SIG // Ex1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDCC
// SIG // AiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAOTh
// SIG // pkzntHIhC3miy9ckeb0O1YLT/e6cBwfSqWxOdcjKNVf2
// SIG // AX9sSuDivbk+F2Az/1xPx2b3lVNxWuJ+Slr+uDZnhUYj
// SIG // DLWNE893MsAQGOhgfWpSg0S3po5GawcU88V29YZQ3MFE
// SIG // yHFcUTE3oAo4bo3t1w/YJlN8OWECesSq/XJprx2rrPY2
// SIG // vjUmZNqYO7oaezOtgFt+jBAcnVL+tuhiJdxqD89d9P6O
// SIG // U8/W7IVWTe/dvI2k45GPsjksUZzpcGkNyjYtcI4xyDUo
// SIG // veO0hyTD4MmPfrVUj9z6BVWYbWg7mka97aSueik3rMvr
// SIG // g0XnRm7KMtXAhjBcTyziYrLNueKNiOSWrAFKu75xqRdb
// SIG // Z2De+JKRHh09/SDPc31BmkZ1zcRfNN0Sidb9pSB9fvzZ
// SIG // nkXftnIv231fgLrbqn427DZM9ituqBJR6L8FA6PRc6ZN
// SIG // N3SUHDSCD/AQ8rdHGO2n6Jl8P0zbr17C89XYcz1DTsEz
// SIG // OUyOArxCaC4Q6oRRRuLRvWoYWmEBc8pnol7XKHYC4jMY
// SIG // ctenIPDC+hIK12NvDMk2ZItboKaDIV1fMHSRlJTYuVD5
// SIG // C4lh8zYGNRiER9vcG9H9stQcxWv2XFJRXRLbJbqvUAV6
// SIG // bMURHXLvjflSxIUXk8A8FdsaN8cIFRg/eKtFtvUeh17a
// SIG // j54WcmnGrnu3tz5q4i6tAgMBAAGjggHdMIIB2TASBgkr
// SIG // BgEEAYI3FQEEBQIDAQABMCMGCSsGAQQBgjcVAgQWBBQq
// SIG // p1L+ZMSavoKRPEY1Kc8Q/y8E7jAdBgNVHQ4EFgQUn6cV
// SIG // XQBeYl2D9OXSZacbUzUZ6XIwXAYDVR0gBFUwUzBRBgwr
// SIG // BgEEAYI3TIN9AQEwQTA/BggrBgEFBQcCARYzaHR0cDov
// SIG // L3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9Eb2NzL1Jl
// SIG // cG9zaXRvcnkuaHRtMBMGA1UdJQQMMAoGCCsGAQUFBwMI
// SIG // MBkGCSsGAQQBgjcUAgQMHgoAUwB1AGIAQwBBMAsGA1Ud
// SIG // DwQEAwIBhjAPBgNVHRMBAf8EBTADAQH/MB8GA1UdIwQY
// SIG // MBaAFNX2VsuP6KJcYmjRPZSQW9fOmhjEMFYGA1UdHwRP
// SIG // ME0wS6BJoEeGRWh0dHA6Ly9jcmwubWljcm9zb2Z0LmNv
// SIG // bS9wa2kvY3JsL3Byb2R1Y3RzL01pY1Jvb0NlckF1dF8y
// SIG // MDEwLTA2LTIzLmNybDBaBggrBgEFBQcBAQROMEwwSgYI
// SIG // KwYBBQUHMAKGPmh0dHA6Ly93d3cubWljcm9zb2Z0LmNv
// SIG // bS9wa2kvY2VydHMvTWljUm9vQ2VyQXV0XzIwMTAtMDYt
// SIG // MjMuY3J0MA0GCSqGSIb3DQEBCwUAA4ICAQCdVX38Kq3h
// SIG // LB9nATEkW+Geckv8qW/qXBS2Pk5HZHixBpOXPTEztTnX
// SIG // wnE2P9pkbHzQdTltuw8x5MKP+2zRoZQYIu7pZmc6U03d
// SIG // mLq2HnjYNi6cqYJWAAOwBb6J6Gngugnue99qb74py27Y
// SIG // P0h1AdkY3m2CDPVtI1TkeFN1JFe53Z/zjj3G82jfZfak
// SIG // Vqr3lbYoVSfQJL1AoL8ZthISEV09J+BAljis9/kpicO8
// SIG // F7BUhUKz/AyeixmJ5/ALaoHCgRlCGVJ1ijbCHcNhcy4s
// SIG // a3tuPywJeBTpkbKpW99Jo3QMvOyRgNI95ko+ZjtPu4b6
// SIG // MhrZlvSP9pEB9s7GdP32THJvEKt1MMU0sHrYUP4KWN1A
// SIG // PMdUbZ1jdEgssU5HLcEUBHG/ZPkkvnNtyo4JvbMBV0lU
// SIG // ZNlz138eW0QBjloZkWsNn6Qo3GcZKCS6OEuabvshVGtq
// SIG // RRFHqfG3rsjoiV5PndLQTHa1V1QJsWkBRH58oWFsc/4K
// SIG // u+xBZj1p/cvBQUl+fpO+y/g75LcVv7TOPqUxUYS8vwLB
// SIG // gqJ7Fx0ViY1w/ue10CgaiQuPNtq6TPmb/wrpNPgkNWcr
// SIG // 4A245oyZ1uEi6vAnQj0llOZ0dFtq0Z4+7X6gMTN9vMvp
// SIG // e784cETRkPHIqzqKOghif9lwY1NNje6CbaUFEMFxBmoQ
// SIG // tB1VM1izoXBm8qGCAsswggI0AgEBMIH4oYHQpIHNMIHK
// SIG // MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3Rv
// SIG // bjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWlj
// SIG // cm9zb2Z0IENvcnBvcmF0aW9uMSUwIwYDVQQLExxNaWNy
// SIG // b3NvZnQgQW1lcmljYSBPcGVyYXRpb25zMSYwJAYDVQQL
// SIG // Ex1UaGFsZXMgVFNTIEVTTjozRTdBLUUzNTktQTI1RDEl
// SIG // MCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2Vy
// SIG // dmljZaIjCgEBMAcGBSsOAwIaAxUAEwa4jWjacbOYU++9
// SIG // 5ydJ7hSCi5iggYMwgYCkfjB8MQswCQYDVQQGEwJVUzET
// SIG // MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
// SIG // bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
// SIG // aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFt
// SIG // cCBQQ0EgMjAxMDANBgkqhkiG9w0BAQUFAAIFAOYQ2SYw
// SIG // IhgPMjAyMjA0MjUxNjI0MDZaGA8yMDIyMDQyNjE2MjQw
// SIG // NlowdDA6BgorBgEEAYRZCgQBMSwwKjAKAgUA5hDZJgIB
// SIG // ADAHAgEAAgILXjAHAgEAAgIRvzAKAgUA5hIqpgIBADA2
// SIG // BgorBgEEAYRZCgQCMSgwJjAMBgorBgEEAYRZCgMCoAow
// SIG // CAIBAAIDB6EgoQowCAIBAAIDAYagMA0GCSqGSIb3DQEB
// SIG // BQUAA4GBAFhO0NTW/1WldAPh6RVNeoQCQbnzxadjroNx
// SIG // edUPcBoFrurxzBPGiUuhryOj9lIpvmx9hQFix3lQEpNd
// SIG // 1UetJcpER+3TQN6iQQsXZfkFSbGfTEpvMtlvJHSReaPO
// SIG // j8xgAk+VmSsRgWaLumtIn05PNAB1fj67emZ+L7pAeC5J
// SIG // FnN5MYIEDTCCBAkCAQEwgZMwfDELMAkGA1UEBhMCVVMx
// SIG // EzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
// SIG // ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
// SIG // dGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3Rh
// SIG // bXAgUENBIDIwMTACEzMAAAGg6buMuw6i0XoAAQAAAaAw
// SIG // DQYJYIZIAWUDBAIBBQCgggFKMBoGCSqGSIb3DQEJAzEN
// SIG // BgsqhkiG9w0BCRABBDAvBgkqhkiG9w0BCQQxIgQgllmF
// SIG // xhG0vQEaVAIRExjLxlmxrzBvSmaWz76GSMcH42swgfoG
// SIG // CyqGSIb3DQEJEAIvMYHqMIHnMIHkMIG9BCAvR4o8aGUE
// SIG // IhIt3REvsx0+svnM6Wiaga5SPaK4g6+00zCBmDCBgKR+
// SIG // MHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
// SIG // dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
// SIG // aWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1p
// SIG // Y3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwAhMzAAAB
// SIG // oOm7jLsOotF6AAEAAAGgMCIEICsUcgbiz0yxxnXnawim
// SIG // T8yMN1Mh7Kyg7gxo0zqhs1iRMA0GCSqGSIb3DQEBCwUA
// SIG // BIICAHLfzAwMOu7LW2dHmwqacGr4H39L3M+x/aJbvZL7
// SIG // 9eZcDFmuS9pPYjIeHdUNJhNxZSORYGbejWHr1/5ZFfLK
// SIG // YkcjhlB+YgAu+eeiSznSt8M1XriWj3AVEGk+Y1Is3J+f
// SIG // K4DCdKCb3kJSFyfeXcekuI1HsuRWXHxqWim1WI2Bwi7t
// SIG // RcMwjla8YOLtZ8pYiUzTpOIuqXItnR1heac8PFweRqX+
// SIG // RbPMGbRhuoOV8aHV2S8ootyR5d6HcDJY6e0OblcdkW7R
// SIG // Ml/nqQizzg2pFdTYseoigBUCCAGcDj4qZZt/6Pdq4N5i
// SIG // Jf1K5MBYpZPwcgLL6JX5udOwqMK/cdbUGQuwLGPS1Kzu
// SIG // WMhoKBwIik7bq1F52xbQVUoNhYJ58s7t6dFhski3KApW
// SIG // Gi0y6Tx4S1lNdid038tZxBZvyT2ZVM8QpNLx4NzKZr77
// SIG // av5NPFyhabQtFqOKybHnSYAYAiQgT1n1XersDwVOkuWU
// SIG // wUet/MwFheZLY1csV/rNj1aWwwGVm0RwpA5bUk+cjAiV
// SIG // xgr7BQIZYnnCwUMPp3L6fJn7VtL6U3R2L0i4OUSS5Ect
// SIG // HJdYPsGc0Z9aJ+r3hMQiDx0/MWx+pq0c8DU7FC7B/YK1
// SIG // r3IcUmJPQdDs/bdx7aBvgMV8VZUAo4p0OSBF7qfVCQy6
// SIG // 5sSHByXAfRFqG6yZJzKVBFiytPdv
// SIG // End signature block
